;;;; AVSOBS -- AVSomat OBServational database
;;;; Copyright (c) 2003-2004 Alexandru Dan Corlan MD PhD
;;;; 

;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.

;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.

;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


(provide "AVSOBS")

;;;; ;; need files: GSVC.lisp allcharts.lisp

(setq *gc-verbose* 'nil)

;;;  global variables to change in order to adapt to your setup

(defvar *avsomat-home* "/home/alexandru/proj/avsomat"
  "Directory were the avsomat source distribution was installed.")

(defvar *report-dir* "/home/alexandru/proj/web/astro.corlan.net/htdocs/obs"
  "Directory in which observation results are stored.")

(defvar *default-obscode* 'CUA)

(defvar *rcp2-dir* (concatenate 'string *avsomat-home* "/rcp2")
  "Directory where recipes are found." )
(setf *rcp2-dir* (concatenate 'string *avsomat-home* "/rcp2"))

(setq *read-default-float-format* 'double-float) ;; needed for julian dates

(defvar *avsomat-home* "/home/alexandru/proj/avsomat")

(defvar *data-dir* (concatenate 'string *avsomat-home*
				"/data"))

(defun from-data (filename)
  "Return the full path where the data file is found."
  (concatenate 'string *data-dir* "/" filename))

;;;; Define and load components of the database

(defvar *constellations*
  '(AND ANT APS AQL AQR ARA ARI AUR BOO CAE
	CAM CAP CAR CAS CEN CEP CET CHA CIR CMA
	CMI CNC COL COM CRA CRB CRT CRU CRV CVN
	CVS CYG DEL DOR DRA EQU ERI FOR GEM GRU
	HER HOR HYA HYI IND LAC LEO LEP LIB LMI
	LUP LYN LYR MEN MIC MON MUS NOR OCT OPH
	ORI PAV PEG PER PHE PIC PSA PSC PUP PYX
	RET SCL SCO SCT SER SEX SGE SGR TAU TEL
	TRA TRI TUC UMA UMI VEL VIR VOL VUL WDC 
	)
  "List of TLA symbols of each constellation"
  )

(setq *constellation-genitives*
  '((AND .  ANDROMEDAE)
    (ANT .  ANTLIAE)
    (APS .  APODIS)
    (AQL .  AQUILAE)
    (AQR .  AQUARII)
    (ARA .  ARAE)
    (ARI .  ARIETIS)
    (AUR .  AURIGAE)
    (BOO .  BOOTIS)
    (CAE .  CAELI)
    (CAM .  CAMELOPARDALIS)
    (CAP .  CAPRICORNI)
    (CAR .  CARINAE)
    (CAS .  CASSIOPEIAE)
    (CEN .  CENTAURI)
    (CEP .  CEPHEI)
    (CET .  CETI)
    (CHA .  CHAMAELEONTIS)
    (CIR .  CIRCINI)
    (CMA .  |CANIS MAJORIS|)
    (CMI .  |CANIS MINORIS|)
    (CNC .  CANCRI)
    (COL .  COLUMBAE)
    (COM .  |COMAE BERENICES|)
    (CRA .  |CORONAE AUSTRINAE|)
    (CRB .  |CORONAE BOREALIS|)
    (CRT .  CRATERIS)
    (CRU .  CRUCIS)
    (CRV .  CORVI)
    (CVN .  |CANUM VENATICORUM|)
    (CYG .  CYGNI)
    (DEL .  DELPHINI)
    (DOR .  DORADUS)
    (DRA .  DRACONIS)
    (EQU .  EQUULEI)
    (ERI .  ERIDANI)
    (FOR .  FORNACIS)
    (GEM .  GEMINORUM)
    (GRU .  GRUIS)
    (HER .  HERCULIS)
    (HOR .  HOROLOGII)
    (HYA .  HYDRAE)
    (HYI .  HYDRI)
    (IND .  INDI)
    (LAC .  LACERTAE)
    (LEO .  LEONIS)
    (LEP .  LEPORIS)
    (LIB .  LIBRAE)
    (LMI .  |LEONIS MINORIS|)
    (LUP .  LUPI)
    (LYN .  LYNCIS)
    (LYR .  LYRAE)
    (MEN .  MENSAE)
    (MIC .  MICROSCOPII)
    (MON .  MONOCEROTIS)
    (MUS .  MUSCAE)
    (NOR .  NORMAE)
    (OCT .  OCTANTIS)
    (OPH .  OPHIUCHI)
    (ORI .  ORIONIS)
    (PAV .  PAVONIS)
    (PEG .  PEGASI)
    (PER .  PERSEI)
    (PHE .  PHOENICIS)
    (PIC .  PICTORIS)
    (PSA .  |PISCIS AUSTRINI|)
    (PSC .  PISCIUM)
    (PUP .  PUPPIS)
    (PYX .  PYXIDIS)
    (RET .  RETICULI)
    (SCL .  SCULPTORIS)
    (SCO .  SCORPII)
    (SCT .  SCUTI)
    (SER .  SERPENTIS)
    (SEX .  SEXTANTIS)
    (SGE .  SAGITTAE)
    (SGR .  SAGITTARII)
    (TAU .  TAURI)
    (TEL .  TELESCOPII)
    (TRA .  |TRIANGULI AUSTRALIS|)
    (TRI .  TRIANGULI)
    (TUC .  TUCANAE)
    (UMA .  |URSAE MAJORIS|)
    (UMI .  |URSAE MINORIS|)
    (VEL .  VELORUM)
    (VIR .  VIRGINIS)
    (VOL .  VOLANTIS)
    (VUL .  VULPECULAE)
    ))

(defun genitive (const)
  "Given CONST, a constellation TLA, returns the genitive of the full constellation name."
  (cdr (assoc const *constellation-genitives*)))

(defun constellation-name (varstar)
  "Given a variable star identifier (like RU-AND) returns the constellation
   TLA which consists in the last 3 letters of the name"
  (let* ((l (length (symbol-name varstar))))
    (intern (subseq (symbol-name varstar) (- l 3)))))

(defun full-name (varstar)
  "Return the full name of the variable star, as a string, given its
   short name as a symbol.
   Example: (full-name 'RU-AND) -> \"RU ANDROMEDAE\""
  (let* ((l (length (symbol-name varstar))))
    (format nil "~A ~A" (subseq (symbol-name varstar) 0 (- l 4))
	    (genitive (constellation-name varstar)))))

(defun full-cgi-name (varstar)
  "Return the full name of the variable star, as a string, given its
   short name as a symbol.
   Example: (full-name 'RU-AND) -> \"RU ANDROMEDAE\""
  (let* ((l (length (symbol-name varstar))))
    (format nil "~A%20~A" (subseq (symbol-name varstar) 0 (- l 4))
	    (genitive (constellation-name varstar)))))


(defvar *with-aavso-charts* nil
  "List of variable star symbols which have aavso charts")

(load (from-data "allcharts.lisp"))
  ;; names of variable stars having aavso charts
  ;; into variable *with-aavso-charts*

(defvar *rcps* nil
  "List of variable star symbols which have rcp charts.")

(load (from-data "rcp.lisp"))
  ;; names of variable stars having rcp charts into *RCPS*
  ;; build this file with the make.rcp.lisp script

(defvar *needmore* nil
  "List of variable star symbols which are currently in need of
   more observations (as LPV variables) according to the aavso bulletin")

(load (from-data "needmore.lisp"))
  ;; stars declared by aavso to be in need of more observations
  ;; the aavso bulletin was processed by hand (in emacs)
  ;; to produce this file; an utility would be welcome

(defvar *vs* nil
  "List of all variable star symbols from the GCVS,\.")

(defun load-gcvs ()

  "Load the lisp version (GCVS.lisp) of the GCVS catalog.
  GCVS.lisp is made from the GCVS III.DAT file with the command gcvs2lisp."

  (setf *vs* nil)
  (with-open-file (fi (from-data "GCVS.lisp"))
		  (do () (nil nil)
		      (let ((la (read fi nil)))
			(when (not la) (return-from load-gcvs))
			(setf *vs* (cons (cadr la) *vs*))
			(setf (get (cadr la) 'ra) (third la))
			(setf (get (cadr la) 'dec) (fourth la))
			(setf (get (cadr la) 'mag) (cons (list (fifth la) (sixth la)) nil))
			(setf (get (cadr la) 'spectype) (eighth la))
			(setf (get (cadr la) 'vartype) (tenth la))
			(setf (get (cadr la) 'aavsoid) (elt la 11))
			))
		  )
  )

(compile 'load-gcvs)

(if (or (not (boundp '*vs*))
	(not *vs*))
    (progn
      (princ (format nil "Loading GCVS... ~%"))
      (load-gcvs)
      (princ (format nil "            ... done!~%"))
      )
  )



;;;; Extract list of stars and information about their reduction/reporting
;;;; status from fits headers and aavso reports, etc

(defun set-reported-obs (obsl)
  
  "Add OBSL, which is a description of a reported observation
   to the list of reported observations of star (CAR OBSL),
   except if it is already there, in which case report a repeated obs."

  (let ((existing-obss (get (car obsl) 'aavso-rep)))
    (if (find (cdr obsl) existing-obss :test #'equal)
	(format *standard-output* "repeated obs: ~A ~A ~A (by ~A)~%"
		(car obsl) (cadr obsl) (caddr obsl) (cadddr obsl))
      (setf (get (car obsl) 'aavso-rep)
	    (cons (cdr obsl) existing-obss)))
    )
  )

(defun clear-all-vs-aavso-rep ()

  "Remove all data about observations reported to aavso from the database."

  (mapcar #'(lambda (x) (setf (get x 'aavso-rep) nil)) *vs*)
  nil)

(defun load-aavso-obslist (fname out-fname)

  "Convert a file downloaded from the `quick look' page at aavso
   into a lisp program which calls the SET-REPORTED-OBS function
   described above, once for each reported observation."

  (with-open-file (fo out-fname :direction :output :if-exists :supersede)
    (with-open-file (fi fname)
	(do ((li "" (read-line fi nil))) ((symbolp li) li)
	    (if (> (length li) 56)
		(format fo "~S~%"
			(list 'set-reported-obs
			      `'(,(intern
				   (substitute #\- #\ 
					       (string-trim " "
							    (subseq li 9 20))))
				 ,(subseq li 32 44)
				 ,(subseq li 45 51)
				 ,(intern (subseq li 53 56)))))
	      )
	    )
	))
  )

(defun extract-fits-header (file-name)

  "Extract the information prezent in the fits header of either a 
   fits file (like one obtained with gcx) or a gzipped fits file named FILE-NAME.
   Return it as an assoc list of keywords and values."

  (let ((fits (make-string-output-stream)))
    (if (string= ".gz" (subseq file-name (- (length file-name) 3) (length file-name)))
	(extensions:run-program "zcat" (list file-name) :output fits)
      (extensions:run-program "cat" (list file-name) :output fits)
      )
    (let ((fidat (get-output-stream-string fits))
	  (filist nil))
      (if (< (length fidat) 10000)
	  (return-from extract-fits-header nil))
      (do* ((fii 0 (1+ fii))
	    (li (subseq fidat 0 80) (subseq fidat (* 80 fii) (* 80 (1+ fii)))))
           ((string= (subseq li 0 4) "END ")
	    filist)
	   (if (position #\/ li)
	       (setq filist (cons (cons (intern (string-trim " " (subseq li 0 8)))
					(string-trim " '" (subseq li 10 (position #\/ li))))
				  filist))
	     (setq filist (cons (cons (intern (string-trim " " (subseq li 0 8)))
				      (string-trim " '" (subseq li 10 80)))
				filist))	       
	     )
	   )
      )
    )
  )

(defun same-obs (obs1 obs2) 
;  (format *standard-output* "~A ~A~%" (assoc 'jdate obs1) (assoc 'jdate obs2))

  "Check if two fits headers represent the same observation."

  (and (equal (cdr (assoc 'object obs1)) (cdr (assoc 'object obs2)))
       (equal (cdr (assoc 'jdate obs1)) (cdr (assoc 'jdate obs2))))
)

(defun object-to-star-name (objname)

  "Convert the object name, as extracted from a fits header
  into a star name (a la RX-AND) as used in lists of variable
  star names in this program."

  (let ((l (length objname)))
    (if (find #\- objname)
	(intern (string-upcase objname))
      (intern (concatenate 'string
			   (string-upcase (subseq objname 0 (- l 3))) "-" 
			   (string-upcase (subseq objname (- l 3) l))))
      )
    )
  )

(defun extract-fits-data (file-name cd-name)

 "Compose a description of a fits header from the fits file FILE-NAME and
  from the volume CD-NAME"

  (let ((fh (extract-fits-header file-name)))
    (setq fh (cons (cons 'VOL cd-name) fh))
    (setq fh (cons (cons 'FITS-NAME file-name) fh))
    )
  )

(defun set-fits-data (fh)

  "Insert the fits header data FH, as produced by EXTRACT-FITS-DATA
   into the database (as an IMAGES property of the starname symbol)."

  (if (assoc 'object fh)
      (let* ((sn (object-to-star-name (cdr (assoc 'object fh)))))
	(if (find fh (get sn 'images) :test #'same-obs)
	    nil ;; observations is already there
	  (setf (get sn 'images) (cons fh (get sn 'images)))
	  )
	sn
	)
    nil)
  )

(defun add-fits-spec (file-name cd-name)

  "Insert the fits header data of the fits file FILE-NAME from volume CD-NAME
  into the database."

  (set-fits-data (extract-fits-data file-name cd-name))
)


(defun add-all-fits-specs (path-set cd-name)

  "Insert the fits header data of all fits files mounted at path PATH-SET
  where the volume CD-NAME is supposed to be found."

  (let ((dir (directory path-set)))
    (mapcar #'(lambda (x)
		(format *standard-output* "~A~%"
			(add-fits-spec (namestring x) cd-name))
		) dir)))

(defun collect-fits-specs (path-set cd-name result-file-name)

  "Given a volume named CD-NAME which is mounted at point PATH-SET
  produce a lisp program, RESULT-FILE-NAME which calls SET-FITS-DATA
  repeatedly to insert into the database the contents of the fits headers
  of all fits files on the CD."

  (let ((fo (open result-file-name :direction :output :if-exists :supersede)))
    (mapcar #'(lambda (x) 
		(format fo "~S~%"
			(list 'set-fits-data
			      (list 'quote
				    (extract-fits-data (namestring x) cd-name))))
		(finish-output fo)
		(format *standard-output* "~A~%" (namestring x))
		)
	    (directory path-set))
    (close fo)
    )
  )

(defun load-list-of-fits ()

  "Load all files of our CDs, as built by COLLECT-FITS-SPECS."

  (mapcar #'(lambda (x) (load (from-data x)))
	  '("d12jul2003.lisp"  "d20sep2003.lisp"  "d27sep2003.lisp"
	    "d31may2003.lisp"  "d19jul2003.lisp"  "d23aug2003.lisp"
	    "d28jun2003.lisp")))

(defun repmag (date replist)
  (dolist (rep replist "      ")
    (if (> 1.1 (abs (- (read-from-string (car rep)) date)))
	(return (cadr rep)))))


(defun report-status-of-star (star)
  (if (get star 'images)
    (format *standard-output* "~A (~A) ~A ~A ~A ~A~%"
	  star (get star 'vartype) (get star 'ra) (get star 'dec)
          (if (listp (get star 'mag))
	      (if (listp (cadar (get star 'mag)))
		  (format nil "~Amag=~A-~A"
			  (caar (get star 'mag))
			  (caadar (get star 'mag))
			  (cadr (cadar (get star 'mag))))
		"mag=?")
	    "mag=?")
          (if (memq star *rcps*) "+rcp+" "")
	  )
    )
    (let ((arep (get star 'aavso-rep)))
      (mapcar
       #'(lambda (x)
	   (let ((jdate (read-from-string (cdr (assoc 'jdate x)))))
	     (format *standard-output* "~20@A ~10,4f ~A~%"
		     (cdr (assoc 'vol x))
		     jdate
		     (repmag jdate arep))))
			       
       (get star 'images))
      )
)

(defun report-status-of-constellation (constell)
  (mapc #'report-status-of-star (in-constellation constell))
  nil)

;;; Remarks accompanying the star in the GCVS

(defun load-gcvs-remarks (&optional (remfile (concatenate 'string *data-dir* "/REM.TXT")))
  (with-open-file (fi remfile)
    (block nil
       (let ((bfr "")
	     (crtsym nil))
	   (do () (nil nil)
	       (let ((la (read-line fi nil)))
		 (when (not la) (return))
		 (if (eql #\  (elt la 0))
		   (setf bfr (concatenate 'string bfr (subseq la 13)))
		   (progn
		     (when crtsym
		       (setf (get crtsym 'gcvs-remark) bfr))
		     (setf crtsym
			(intern (string-upcase
				 (concatenate 'string
				  (string-trim " " (subseq la 0 5))
				  "-"
				  (string-trim " " (subseq la 6 13))))
				)
			)
		     (setf bfr (subseq la 13))
		     )
		   )
		 )))
       )
    )
)

(compile 'load-gcvs-remarks)

(load-gcvs-remarks)

;;; load links and other data pertaining to the stars
;;; declarations using the macros below are stored into
;;; databases like data/linkbase.lisp

(defmacro setfor (&rest frm)
  nil)

(defmacro setlink (&rest ll)
  `(setf (get ',(car ll) 'links)
	 (cons ',(cdr ll) (get ',(car ll) 'links))))

(defmacro setlinks (&rest ll)
  `(dolist (sta ',(car ll))
     (setf (get sta 'links)
	   (cons ',(cdr ll) (get sta 'links))))
  )

(defun read-assoc-component (item assoc-list)
  "Read the string which is associated with the item in the list."
  (with-input-from-string (s (cdr (assoc item assoc-list))) (read s)))

(defun diaskys-line (filename)
  "Produce information regarding FILENAME in the form needed by
   the Distributed Involuntary Amateur Sky Survey."
  (let* ((fh (extract-fits-header filename))
	 (crval1 (read-assoc-component  'crval1 fh))
	 (crval2 (read-assoc-component  'crval2 fh))
	 (crpix1 (read-assoc-component  'crpix1 fh))
	 (crpix2 (read-assoc-component  'crpix2 fh))
	 (cdelt1 (read-assoc-component  'cdelt1 fh))
	 (cdelt2 (read-assoc-component  'cdelt2 fh))
	 (naxis1 (read-assoc-component  'naxis1 fh))
	 (naxis2 (read-assoc-component  'naxis2 fh))
	 (jdate  (read-assoc-component  'jdate fh))
	 (exptime  (read-assoc-component  'exptime fh))
	 (ra1 (- crval1 (/ (* cdelt1 crpix1) (cos (* (/ 3.14 180.0) crval2)))))
	 (ra2 (+ crval1 (/ (* cdelt1 (- naxis1 crpix1)) (cos (* (/ 3.14 180.0) crval2)))))
	 (dec1 (- crval2 (* cdelt2 crpix2)))
	 (dec2 (+ crval2 (* cdelt2 (- naxis2 crpix2))))
	  )
    ;          MINRA   MAXRA     MINDEC    MAXDEC    JDATE     EXPTIME  FILENAME
    (format *standard-output*
	    "~12,6,,,F ~12,6,,,F ~12,6,,,F ~12,6,,,F ~12,4,,,F ~8,2,,,F ~A~%"
	    (min ra1 ra2)
	    (max ra1 ra2)
	    (min dec1 dec2)
	    (max dec1 dec2)
	    jdate exptime filename)
    )
  )

(compile 'diaskys-line)

(defun diaskys-dir (path-set)
  "Output the data for the use of the Distributed Involuntary Amateur SKY Survey
   for all fits file in the directory PATHSET"

  (let ((dir (directory path-set)))
    (mapcar #'(lambda (x)
		(with-simple-restart (nil "Ignore image, go to the next")
				     (diaskys-line (namestring x))
				     )
		) dir)))

;;; must load at most once ...

(require "LINKBASE" (concatenate 'string *data-dir* "/linkbase.lisp")
	 )

;;; ... or execute the following

(defun reload-linkbase ()
  (mapcar #'(lambda (x) (setf (get x 'links) nil))
	  *vs*)
  (load (concatenate 'string *data-dir* "/linkbase.lisp"))
)

