// fieldcheck -- check that a fits image corresponds to the star pattern in a recipe
// Copyright (c) 2003 Alexandru Dan Corlan (http://dan.corlan.net)
//

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "libnut.h"
#include "../ccd/ccd.h"

static float xs[10000], ys[10000], ms[10000];
static float catxs[10000], catys[10000], catms[10000];
static int matches[10000];
static float romega, rdx, rdy, rampx, rampy, rdmag;
static int nc, ns;



static void rot_vect(float *x, float *y, double a)
{
	float nx, ny;

	nx = *x * cos(a) - *y * sin(a);
	ny = *x * sin(a) + *y * cos(a);
	*x = nx; *y = ny;
}


//extern int tan_xypix(double xpos, double ypos,
//		     double xref, double yref, double xrefpix, double yrefpix,
//		     double xinc, double yinc, double rot,
//		     double *xpix, double *ypix);

int field_check(struct ccd_frame *fr,  	// image frame to check
		float snr,             	// signal to noise ratio for src extraction
		float rsearch,         	// radius to search around image center
		struct wcs *fim)  	// computed wcs for frame
{
  float plate_center_ra;
  float plate_center_dec;
  float deg_per_pix;

  static int regs[10000], ids[10000];
  double tempx, tempy;
  int i,j, ret;

  struct sources *src;

  plate_center_ra = fr->fim.xref;
  plate_center_dec = fr->fim.yref;
  deg_per_pix = fr->fim.xinc;

  nc = getgsc(plate_center_ra, plate_center_dec, rsearch, 15.0, regs, ids,
	      catxs, catys, catms, 10000);
  for(i=1, j=1; i<nc; i++) {
    while((regs[i]==regs[i-1]) && (ids[i]==ids[i-1]))
      i++;
    if (i>j) { // there are many duplicates in gsc
      regs[j] = regs[i];
      ids[j] = ids[i];
      catxs[j] = catxs[i];
      catys[j] = catxs[i];
      catms[j] = catms[i];
    };
    j++;
  }
  nc = j;

  for (i=0; i<nc; i++) {
    xypix(catxs[i],catys[i],plate_center_ra,plate_center_dec, 0.0, 0.0,
	  deg_per_pix,deg_per_pix, 0.0, "-TAN", &tempx, &tempy);
    catxs[i] = tempx;
    catys[i] = tempy;
  }

  src = new_sources(100);

  if (src == NULL) {
    exit(-1);
  }
  ns = extract_stars(fr, NULL, (double)0.0, snr, src);

  for(i=0; i<ns; i++) {
    xs[i] = src->s[i].x;
    ys[i] = src->s[i].y;
    ms[i] = flux_to_absmag(src->s[i].flux);
  }

  free_sources(src);

  ret = starmatch(xs, ys, ms, ns,
		   catxs, catys, catms, nc,
		   matches, &romega, &rdx, &rdy, &rampx, &rampy, &rdmag);

  if (ret < 3) {
	  err_printf("only matched %d stars, need at least 3\n", ret);
	  return 0;
  }

  d3_printf("omega:%.4f dx:%.1f dy:%.1f ampx:%.4f ampy:%.4f\n",
	    romega, rdx, rdy, rampx, rampy); 


  //  rot_vect(&rdx, &rdy, -romega);

  //  rdx *= rampx;
  //  rdy *= rampy;

// init new wcs
  memcpy(fim, &(fr->fim), sizeof(struct wcs));
// compute the ra/dec of the new reference pixel
  worldpos(2 * fim->xrefpix - rdx + 1, 2 * fim->yrefpix - rdy , fr->fim.xref, fr->fim.yref,
	   fim->xrefpix, fim->yrefpix, fim->xinc, fim->yinc, fim->rot,
	   "-TAN", &(fim->xref), &(fim->yref));
// add rotation and scale
  fim->rot -= 180 / PI * romega;
  fim->xinc *= sqrt(rampx * rampy);
  fim->yinc *= sqrt(rampx * rampy);
  return ret;
}

